/**
 * Very spaghetti. But works. =)
 */


import java.sql.*;
import java.util.*;
import java.util.concurrent.atomic.AtomicLong;

import sun.misc.Perf.GetPerfAction;


import org.apache.derby.shared.common.sanity.SanityManager;

class Select {

    // "Globale" variablar.
    static int numWorkers = 8;
    static int numTxns = 10000;
    static int scaleFactor = 100;
    static int warmup = numTxns;
    static boolean initializeDatabase = false;
    static boolean cleanDatabase = false;
    static boolean doRun = true;
    static boolean latches = false, owntable = false;
    static String dbName = "select";
    static String tableName = "foo";

    static Worker.WorkerType runtype = Worker.WorkerType.SELECT;

    static void usage(String message) {
        if(message == null) {
            System.err.println("Simple workload generator for Derby.\n");
        } else {
            System.err.printf("Error: %s\n", message );
        }
        System.err.println("Command line usage:");
        System.err.println("    -h: This text");
        System.err.println("    -d <dbnamn>: Database name. (default " + dbName + ")");
        System.err.println("    -drop: drop tables before init.");
        System.err.println("    -init: Initialize database before test run");
        System.err.println("    -justinit: Initialize database, no test run..");
        System.err.println("    -update: Run UPDATE transactions (default is SELECT)");
        System.err.println("    -delete: Run DELETE transactions");
        System.err.println("    -latches: Output latch wait statistics");
        System.err.println("    -owntable: Run each worker thread on it's own table. (Init first!)");
        System.err.println("    -s <N>: Scaling factor: Number of thousands of row in the database. (default " + scaleFactor + ")");
        System.err.println("    -n <N>: Number of transactions per worker thread (default " + numTxns + ")");
        System.err.println("    -t <N>: Number of worker threads (default " + numWorkers + ")");
        System.err.println("    -w <N>: Antall transaksjoner for warmup-tråd, 0 = disable (default " + numTxns + ")");
        System.exit(1);
    }

    static void parseArgs(String[] argarray) {
        ListIterator<String> args = Arrays.asList(argarray).listIterator();

        while(args.hasNext()) {
            String arg = args.next();
            String next = null;
            if ("-d".equals(arg)) {
                try {
                    next = args.next();
                    dbName = next;
                } catch (NoSuchElementException e) {
                    usage("Missing parameter for -s");
                }
            } else if ("-drop".equals(arg)) {
                cleanDatabase = true;
            } else if("-h".equals(arg)) {
                usage(null);
            } else if ("-init".equals(arg)) {
                initializeDatabase = true;
            } else if ("-justinit".equals(arg)) {
                initializeDatabase = true;
                doRun = false;
            } else if ("-update".equals(arg)) {
                runtype = Worker.WorkerType.UPDATE;
            } else if ("-delete".equals(arg)) {
                runtype = Worker.WorkerType.DELETE;
            } else if ("-latches".equals(arg)) {
                latches = true;
            } else if ("-owntable".equals(arg)) {
                owntable = true;
            } else if ("-n".equals(arg)) {
                try {
                    next = args.next();
                    numTxns = Integer.parseInt(next);
                } catch (NoSuchElementException e) {
                    usage("Missing parameter for -n");
                    usage("Manglande parameter til -n");
                } catch (NumberFormatException e) {
                    usage("Invalid parameter for -n: " + next);
                }
            } else if ("-s".equals(arg)) {
                try {
                    next = args.next();
                    scaleFactor = Integer.parseInt(next);
                } catch (NoSuchElementException e) {
                    usage("Missing parameter for -s");
                } catch (NumberFormatException e) {
                    usage("Invalid parameter for -s: " + next);
                }
            } else if ("-t".equals(arg)) {
                try {
                    next = args.next();
                    numWorkers = Integer.parseInt(next);
                } catch (NoSuchElementException e) {
                    usage("Missing parameter for -t");
                } catch (NumberFormatException e) {
                    usage("Invalid parameter for -t: " + next);
                }
            } else if ("-w".equals(arg)) {
                try {
                    next = args.next();
                    warmup = Integer.parseInt(next);
                } catch (NoSuchElementException e) {
                    usage("Manglande parameter til -w");
                } catch (NumberFormatException e) {
                    usage("Ugyldig parameter til -w: " + next);
                }
            }
        }

    }

    /**
     * Determines if a table exists in the database c is connected to.
     *
     * @param c Connection to a database
     * @param tablename Name of the table to look for
     * @return true if the table exists, false otherwise.
     * @throws SQLException If an exception is thrown by the queries in this method,
     * it is simply passed on up.
     */
    static boolean hasTable(Connection c, String tablename) throws SQLException {
        DatabaseMetaData metadata = c.getMetaData();

        ResultSet tables = metadata.getTables(null, null, tablename.toUpperCase(), null);

        return tables.next(); // If there's at least one row in the resultset, the table exists.
    }

    /**
     * Initialize a database with test data. Optionally destroys existing table before
     * creating a new one and filling it with appropriate amounts of data according to
     * the scale factor specified.
     */
    static void initDB(String tableName, String copyTable) {
        Connection c;
        try {
            c = DriverManager.getConnection("jdbc:derby:" + dbName + ";create=true");
            c.setAutoCommit(false);
        } catch (SQLException e) {
            System.err.println("Could not connect to/create database: " + e.getMessage());
            System.exit(1);
            /* NOTREACHED */
            return; // Convinces the compiler that c will always be initialized.
        }

        try {

            /**
             * First we initialize the table
             */
            createTable(tableName, c);

            if(copyTable == null) {
                insertRows(c, tableName);
            } else {
                Statement s = c.createStatement();
                s.executeUpdate("INSERT INTO " + tableName + " ( SELECT * FROM " + copyTable + ")");
                c.commit();
            }

            c.close();


        } catch (SQLException e) {
            System.err.println("SQLException: " + e.getMessage());
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }

    }

    private static void insertRows(Connection c, String tableName) throws SQLException {
        PreparedStatement insert = c.prepareStatement("INSERT INTO " + tableName + " (a,b) VALUES(?, ?)");
        byte[] foo = "".getBytes();
        insert.setBytes(2, foo);

        /* We don't bother with inserting rows already in the table - and we assume everything up to MAX(a) is
  * filled. */
        String maxSQL = "SELECT MAX(A) FROM " + tableName;
        Statement once = c.createStatement();

        ResultSet getmax = once.executeQuery(maxSQL);
        getmax.next();
        int largestrow = getmax.getInt(1);


        int rowsInserted = 0;
        for(int i = largestrow; i< scaleFactor*1000; i++) {
            insert.setInt(1, i);

            //System.out.println("Inserting row " + i);
            try {
                insert.executeUpdate();
                rowsInserted++;
            } catch (SQLException e) {
                //System.err.println("Failed to insert row " + i + ": " + e.getMessage());
                // God old catch(foo) {} =)
            }
            if((rowsInserted % 100) == 0) {
                c.commit();
                System.err.println("Committed at " + rowsInserted + " rows inserted");
            }

        }
        System.out.printf("Inserted %d rows\n", rowsInserted);
        c.commit();
    }

    private static void createTable(String tableName, Connection c) throws SQLException {
        String createSQL = "CREATE TABLE " + tableName +" (A INTEGER NOT NULL, \n" +
                            "B CHAR(96) FOR BIT DATA NOT NULL, \n" +
                            "PRIMARY KEY(A))";
        String dropSQL = "DROP TABLE " + tableName;
        Statement once = c.createStatement();

        if(cleanDatabase && hasTable(c, tableName)) {
                once.executeUpdate(dropSQL);
        }

        if(!hasTable(c, tableName)) {
            once.executeUpdate(createSQL);
        }
    }

    static void warmup(String tableName, int warmupThreads) {
        String selectAllSQL = "SELECT * FROM ";
        String selectIndexSQL = "SELECT a FROM ";

        Connection c = null;
        PreparedStatement statement = null;
        try {
            c = DriverManager.getConnection("jdbc:derby:" + dbName);
            c.setReadOnly(true);
            c.setTransactionIsolation(c.TRANSACTION_READ_COMMITTED);
            // kjører warmup-worker
            if(warmupThreads == 1) {
                System.err.println("Warming up cache by selecting whole table.");
                statement = c.prepareStatement(selectAllSQL + tableName);
                ResultSet rs = statement.executeQuery();
                // touch resultset
                while (rs.next()) {
                    rs.getInt(1);
                }
                System.err.println("Warming up with " + warmup + " transactions.");
                Worker warmupWorker = Worker.newWorker(runtype, null,"WarmupWorker", dbName, tableName, scaleFactor, warmup);
                warmupWorker.start();
                warmupWorker.join();
            } else if(warmupThreads > 1) {
                for(int i = 0; i < warmupThreads; i++) {
                    String workerTable = tableName + i;
                    System.err.println("Warming up cache by selecting whole " + workerTable);
                    statement = c.prepareStatement(selectIndexSQL + workerTable);
                    ResultSet rs = statement.executeQuery();
                    // touch resultset
                    while (rs.next()) {
                        rs.getInt(1);
                    }
                    statement = c.prepareStatement(selectAllSQL + workerTable);
                    rs = statement.executeQuery();
                    while (rs.next()) {
                        rs.getInt(1);
                    }
                }
                System.err.println("Warming up with " + warmupThreads + " threads each doing " + warmup + " transactions.");
                ThreadGroup warmupGroup = new ThreadGroup("worker threads");
                Vector<Worker> warmupWorkers = new Vector<Worker>(warmupThreads);
                for(int i = 0; i < warmupThreads; i++) {
                    String workerTable = tableName + i;
                    warmupWorkers.add(Worker.newWorker(runtype, warmupGroup, "Warmer" + i, dbName, workerTable, scaleFactor, warmup));
                }
                for(Worker w: warmupWorkers) {
                   w.start();
               }

                // Make sure all workers are finished
                for(Worker w: warmupWorkers) {
                    try {
                        w.join();
                    } catch (InterruptedException e) {
                        // Yeah, whatever.
                    }
                }
            } else {
                System.err.println("Excuse me? " + warmupThreads + " warmup threads?");
            }
            System.err.println();
            
        } catch (SQLException e) {
            System.err.println("Feil under kjøring av warmup()");
            e.printStackTrace();
            System.exit(0);
        } catch (java.lang.InterruptedException e) {
            System.err.println("InterruptedException i warmup()! WTF!");
        }
    }

    public static void main(String[] args) {
        try {
            Class.forName("org.apache.derby.jdbc.EmbeddedDriver");
        } catch (ClassNotFoundException e) {
            System.err.println("Derby missing in $CLASSPATH. Fix.");
        }

        parseArgs(args);

        if(initializeDatabase) {
            if(owntable) {
                for(int i = 0; i < numWorkers; i++) {
                    initDB(tableName + i, tableName);
                }
            } else {
                initDB(tableName, null);
            }
        }

        if (doRun && warmup != 0) {
            warmup(tableName, owntable ? numWorkers : 1 );
        }
        if(doRun) {
            System.out.printf("Preparing for run with %d worker threads, scaling factor %d, %d transactions per worker.%n",
                    numWorkers, scaleFactor, numTxns);
            ThreadGroup workergroup = new ThreadGroup("worker threads");
            Vector<Worker> workers = new Vector<Worker>(numWorkers);
            for(int i = 0; i < numWorkers; i++) {
                String workerTable = owntable ? tableName + i : tableName;
                workers.add(Worker.newWorker(runtype, workergroup, "Worker" + i, dbName, workerTable, scaleFactor, numTxns));
            }

            //SanityManager.LOCKTRACE = false;
            //try {Thread.sleep(1);} catch (Exception e) {};
            long startTime = System.currentTimeMillis();
            System.out.printf("Run starts at %TT%n", startTime);


            for(Worker w: workers) {
                w.start();
            }

            /* Wait for most workers to finish - if only one is still left we'll
             * wait for it below.
             */
            while(workergroup.activeCount() > 1) {
                try {
                    Thread.sleep(100);
                } catch (InterruptedException e) {
                    // Yeah, whatever.
                }
            }

            // Make sure all workers are finished
            for(Worker w: workers) {
                try {
                    w.join();
                } catch (InterruptedException e) {
                    // Yeah, whatever.
                }
            }

            //SanityManager.LOCKTRACE = false;
            System.out.print("\n\n"); // Blank line before we print the report

            long totalTime = System.currentTimeMillis() - startTime;
            System.out.printf("Run finished, took %d:%d:%d.%d%n%n",
                                totalTime / (60*60*1000),
                                (totalTime / (60 * 1000)) % 60,
                                (totalTime / 1000 ) % 60,
                                totalTime % 1000);


            // Collect statistics, aggregate and report.
            Vector<TransactionStats> allstats = new Vector<TransactionStats>();
            for(Worker w: workers) {
                allstats.add(w.getStats());
            }

            /**
             * Utility class for keeping track of a page latch's
             * total wait times, and for sorting them.
             */
            class TimedPage implements Comparable {
                Long time; String page;
                TimedPage(Long time, String page) {
                    this.time = time;this.page = page;
                }

                /**
                 * Merk at ein taktisk minus nedanfor betyr omvendt
                 * sortering (høgast fyrst)
                 * @param o
                 * @return
                 */
                public int compareTo(Object o) {
                    return -time.compareTo(((TimedPage)o).time);
                }
                public String toString() {
                    return page + ": " + time;
                }
            }

            TransactionStats agg = TransactionStats.aggregateStats(allstats);
            agg.printReport();
            System.out.printf("Total throughput: %d txns/second.%n", (agg.txCount*1000) / totalTime);

            if(latches) {
                long totalpagewaits = 0;
                List<TimedPage> pages = new ArrayList<TimedPage>();

                for(Map.Entry<String, AtomicLong> page : SanityManager.pagewaits.entrySet()) {
                    long time; // no see?
                    totalpagewaits += time = page.getValue().longValue();
                    pages.add(new TimedPage(time, page.getKey()));
                }
                TimedPage[] tmparray = {};
                TimedPage[] pgarray = pages.toArray(tmparray); // (TimedPage[]) 
                Arrays.sort(pgarray);
                pages = Arrays.asList(pgarray);

                System.out.printf("Total time spent waiting for latches: %d milliseconds.%nTop latches:%n", totalpagewaits/1000L);
                int i = 0;
                for(TimedPage page: pages) {
                    System.out.println(page);
                    if(++i > 25) break;
                }
            }

        }
    }

}
