/**
 * Created by IntelliJ IDEA.
 * User: andersmo
 * Date: 31/08/2006
 * Time: 10:02:00
 * To change this template use File | Settings | File Templates.
 */
import java.util.Collection;

/**
 * Tek vare på statistikk for ein arbeidstråd som gjer transaksjonar. Ikkje trådsikker, bruk ein per tråd.
 */
class TransactionStats {
    // Antal transaksjonar gjennomført
    long txCount = 0;
    // Antal feila transaksjonar
    long txFailed = 0;
    // Total tid forbrukt på transaksjonar
    long totalTime = 0;
    // Lengste tid på ein transaksjon
    long maxTime = Long.MIN_VALUE;
    // Kortaste tid brukt på ein transaksjon
    long minTime = Long.MAX_VALUE;

    // Starttidspunkt for siste transaksjon
    private long lastStart;
    // Identifiserer denne statistikksamlaren ved rapportering
    private String ident;

    /**
     * Sett opp eit TransactionStats-objekt med ein identifikator som kan brukast til rapportering seinare.
     * @param identifier Brukarvald streng, blir skrivi ut av report.
     */
    public TransactionStats(String identifier) {
        ident = identifier;
    }

    /**
     * Aggregerer statistikk frå fleire TransactionStats-objekt. Summerer txCount og txFailed, summerer
     * totalTime (for å få gjennomsnittleg transaksjonstid for alle), finn største maxTime i alle og minste minTime
     * i alle.
     * @param statsObjects Ei samling TransactionStats
     * @return Aggregert statistikk.
     */
    public static TransactionStats aggregateStats(Collection<TransactionStats> statsObjects) {
        TransactionStats agg = new TransactionStats("aggregated");
        for(TransactionStats stat: statsObjects) {
            agg.txCount += stat.txCount;
            agg.txFailed += stat.txFailed;
            agg.totalTime += stat.totalTime;
            agg.minTime = (stat.minTime < agg.minTime)? stat.minTime : agg.minTime;
            agg.maxTime = (stat.maxTime > agg.maxTime)? stat.maxTime : agg.maxTime;
        }
        return agg;
    }

    /**
     * Kall denne når du startar ein transaksjon, den initierer tidtakinga.
     */
    public void txStart() {
        lastStart = System.nanoTime();
    }

    /**
     * Kall denne når transaksjonen er ferdig - enten den feila eller ikkje.
     * @param failed Send inn true her om transaksjonen aborta elns.
     */
    public void txDone(boolean failed) {
        // deler på ein mill for å gå frå nano- til millisekund.
        long elapsed = (System.nanoTime() - lastStart) / 1000;
        txCount++;
        if(failed) txFailed++;
        totalTime += elapsed;
        maxTime = (elapsed > maxTime) ? elapsed : maxTime;
        minTime = (elapsed < minTime) ? elapsed : minTime;
    }

    public void printReport() {
        System.out.printf("Transaction statistics report: '%s'%n", ident);

        System.out.printf("Total txs: %d, %d failed.%n", txCount, txFailed);
        System.out.printf("Average response time: %f ms%n", (double)totalTime / (txCount * 1000));
        System.out.printf("Maximum/minimum response time: %d us/%d us %n", maxTime, minTime);
    }

}
