import java.sql.*;

/**
 * Created by IntelliJ IDEA.
 * User: andersmo
 * Date: 31/08/2006
 * Time: 10:19:48
 * To change this template use File | Settings | File Templates.
 */
public abstract class Worker extends Thread {
    protected Connection c;
    protected PreparedStatement myStatement;

    public enum WorkerType { SELECT, UPDATE, DELETE };

    // String selectSQL;
    String dbName;
    int scaleFactor;
    long runTxns;

    TransactionStats myStats;

    /**
     * Prepares a worker thread to do a lot of SELECTing
     * @param group ThreadGroup for the thread to live in.
     * @param name A name to refer to the thread by.
     * @param dbName Database name to connect to
     * @param tableName Table name to mess with
     * @param scaleFactor Database scaling factor - how large is it supposed to be?
     * @param numTxns Number of transactions to run.
     */
    protected Worker(ThreadGroup group, String name, String dbName, String tableName, int scaleFactor, long numTxns) {
        super(group, name);

        this.dbName = dbName;

        this.scaleFactor = scaleFactor;
        this.runTxns = numTxns;

        try {
            c = DriverManager.getConnection("jdbc:derby:" + dbName);
            //c.setReadOnly(true);                                        // Read-only workload.
            c.setReadOnly(false);
            c.setTransactionIsolation(c.TRANSACTION_READ_COMMITTED);    // For obviousness' sake.
            myStatement = prepareWorkStatement(tableName);
        } catch (SQLException e) {
            System.err.println("WTF: SQLException in worker thread setup: " + e.getMessage());
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            System.exit(1);
        }

        myStats = new TransactionStats(name);
    }

    /**
     * Factorymetode for å lage workere som gjer ulike ting.
     * @param type Skal tråden gjere select, update eller delete?
     * @param group ThreadGroup for the thread to live in.
     * @param name A name to refer to the thread by.
     * @param dbName Database name to connect to
     * @param tableName Table name to mess with
     * @param scaleFactor Database scaling factor - how large is it supposed to be?
     * @param numTxns Number of transactions to run.
     * @return workertråd, klar til bruk.
     */
    public static Worker newWorker(WorkerType type, ThreadGroup group, String name, String dbName, String tableName, int scaleFactor, long numTxns) {
        switch (type) {
            case SELECT: return new SelectWorker(group, name, dbName, tableName, scaleFactor, numTxns);
            case UPDATE: return new UpdateWorker(group, name, dbName, tableName, scaleFactor, numTxns);
            case DELETE: return new DeleteWorker(group, name, dbName, tableName, scaleFactor, numTxns);
            default: return null;
        }
    }

    /**
     * Worker thread - this method actually does the work.
     */
    public void run() {
        ResultSet rs;
        int updateResult;
        int randomRow;
        System.out.printf(" +%s", this.getName());
        for(long i = 0; i < runTxns; i++) {
            try {
                randomRow = (int) (Math.random() * scaleFactor * 1000);
                setupStatement(randomRow);
                myStats.txStart();
                myStats.txDone(!(execStatement(randomRow)));
            } catch (SQLException e) {
                myStats.txDone(true); // ups, feila visst. jaja.
                //e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
            }
        }
        try {
            c.close();
        } catch (SQLException e) {
            // Whaat?
        }
        System.out.printf(" -%s", this.getName());
    }

    abstract PreparedStatement prepareWorkStatement(String tableName) throws SQLException;
    abstract void setupStatement(int rowId) throws SQLException;
    abstract boolean execStatement(int rowId) throws SQLException;
    
    public TransactionStats getStats() {
        return myStats;
    }
}
